package provide classSupport 1.0
set rev(skeleton.tcl) {$Header: /home/cvs/EditTable/classSupport-1.0.tm,v 1.1 2012/12/29 22:09:42 clif Exp $}
#####################################################################
;# NAME:   classSupport
;# AUTHOR: Clif Flynt
;# DATE:   8/2012
;# DESC:   Provide support for class methods and variables.
;#
;# PARAMETERS:  
;#  See below.
;#
#####################################################################
# This code is written and copyrighted by Clif Flynt, 2012
# Stolen and modified from code posted by Donal Fellows to
# wiki.tcl.tk
#
# It is licenced for public use with these terms:
#
#  This licence agreement must remain intact with the code.
# 
#  You may use this code freely for non-commercial purposes.
#
#  You may examine and modify the code to your heart's content.
#
#  You may not claim to have authored the code (aside from modifications
#  you may have made.)
#
#  You may not distribute the code without this license agreement
#
#  You may not sell this code or distribute it with a commercial product
#      without arrangements with Noumena Corporation.
#
#  Maintenance, extension, modification, etc can be performed by:
#
#   Noumena Corporation
#   8888 Black Pine Ln
#   Whitmore Lake, MI  48189
#
#   Contact: clif@noucorp.com
#
#####################################################################

# Create classvar proc if it doesn't exist.
if {[info proc ::oo::Helpers::classvar] eq ""} {

  ################################################################
  # proc ::oo::Helpers::classvar {args}--
  #    Create a variable associated with a class
  # Arguments
  #   args - list of variables to be associated with a class
  # 
  # Results
  #   Modifies the namespace associated with the class 
  # 
  proc ::oo::Helpers::classvar {args} {
      # Get reference to class’s namespace
      set ns [info object namespace [uplevel 1 {self class}]]
      # Double up the list of varnames
      foreach v $args {
        uplevel 1 namespace upvar $ns $v $v
      }
  }
}

# Create classmethod proc if it doesn't exist.
if {[info proc ::oo::define::classmethod] eq ""} {
  ################################################################
  # proc ::oo::define::classmethod {name {args ""} {body ""}}--
  #    Create a method associated with class, rather than object
  # Arguments
  #   name - Name of the method
  #   args - Argument list, following Tcl conventions
  #   body - Body of procedure
  # 
  # Results
  #   Modifies the class definitions
  # 
  proc ::oo::define::classmethod {name {args ""} {body ""}} {
     # Create the method on the class if the caller gave
     # arguments and body
     if {[llength [info level 0]] == 4} {
          uplevel 1 [list self method $name $args $body]
     }
     # Get the name of the class being defined
     set cls [lindex [info level -1] 1]
     # Make connection to private class "my" command by
     # forwarding
     uplevel forward $name [info object namespace $cls]::my $name
  }
}

