#####################################################################
# This code is written and copyrighted by Clif Flynt, 2017
#
# It is licenced for public use with these terms:
#
#  This licence agreement must remain intact with the code.
# 
#  You may use this code freely for commercial or non-commercial purposes.
#
#  You may examine and modify the code to your heart's content.
#
#  You may not claim to have authored the code (aside from modifications
#  you may have made.)
#
#  You may not distribute the code without this license agreement
#
#  Maintenance, extension, modification, etc can be performed by:
#
#   Noumena Corporation
#   8888 Black Pine Ln
#   Whitmore Lake, MI  48189
#
#   Contact: clif@noucorp.com
#
#####################################################################

catch {SQLITE3_support destroy}

 oo::class create SQLITE3_support {

#
# Methods the DB class must implement.  
#

################################################################
#   method init {dbEngine args}--
#    Initialize connection to DB
# Arguments
#   dbEngine	Type of engine 
#   args	Arguments the engine requires
# 
# Results
#   Opens a DB connection
# 
  method init {dbEngine args} {
    # Args: dbName dbFileName
    variable State
    sqlite3 [self]_db {*}$args
    set State(db) [self]_db 
  }
  
################################################################
#   method closeDB {}--
#    Close the connection
# Arguments
#   NONE
# 
# Results
#   DB connection is closed
# 
  method closeDB {} {
    variable State
    $State(db) close
  }

################################################################
#   method getTables {}--
#    Return a list of tables in the DB
# Arguments
#   NONE
# 
# Results
#   No side effects
# 
  method getTables {} {
    variable State
    set rtn [$State(db) eval "SELECT DISTINCT name FROM sqlite_master WHERE \
        type = 'table'"]
    return $rtn
  }

################################################################
#   method getValueByRef {tbl fld primary value}--
#    Return the value in a field based on primary key and value
# Arguments
#   tbl		Name of table
#   fld		Name of field to return a value for
#   primary	Name of primary field 
#   value	Value for primary.
# 
# Results
#   No side effects
# 
  method getValueByRef {tbl fld primary value} {
    variable State
# puts "GetValueByRef tbl $tbl fld $fld primary $primary value $value "
    set rtn [$State(db) eval "SELECT $fld FROM $tbl WHERE $primary = :value"]
    return {*}$rtn
  }
  
################################################################
#   method populateWithFwdRef {search}--
#    Populate for fields that reference other fields
# Arguments
#   search	SQL search terms
# 
# Results
#   Update State(rows)
# 
  method populateWithFwdRef {search} {
    variable State
    set State(rows) {}
# puts "POP W/FWDREF: ..$search.."
    set rtn ""
    set sch [my getSchema $State(table) 0]
    $State(db) eval "SELECT * FROM $State(table) WHERE $search" xx {
# parray xx
      catch {unset xx(*)}
      set row [array get xx]
      set tblList {}
      foreach no $State(refObjs) {
        set noTbl  [$no config -table]
	if {[lsearch $tblList $noTbl] >= 0} {continue}
	lappend tblList $noTbl
        set noPrim [$no getPrimary $noTbl]
        set ll [my getSchemaLineRef $sch $noTbl]
# puts "noTBL: $noTbl PRIM: $noPrim"
# puts "SCH: $sch"
# puts "LL: ..$ll.."
        set p [string first " " $ll]
        set noFld [string trim [string range $ll 0 $p]]
        # puts "NOFLD : $noFld"
	set valDict [$no populateBySearch "$noPrim = $xx($noFld)"]
        # puts "searchRTN: $noPrim == $xx($noFld) -- $valDict"
	foreach {k v} {*}$valDict {
# puts "K: ..$k.. V: ..$v.."
	  lappend row $noTbl.$k $v
	}
      }
      lappend State(rows) $row
    }
    if {$State(refSel) ne ""} {
      $State(refSel) configure -from 0 -to [expr {[llength $State(rows)]-1}]
      $State(refSel) set 3
      $State(refSel) set 0
    }
  }

################################################################
#   method populateBySearch {search}--
#    Return values from a search
# Arguments
#   search	SQL test phrase
# 
# Results
#   Update State(rows)
# 
  method populateBySearch {search} {
    variable State
    set rtn ""
#  puts "PopSear: [self] -- SELECT * FROM $State(table) WHERE $search"
    $State(db) eval "SELECT * FROM $State(table) WHERE $search" xx {
      set row {}
      foreach in [array names xx] {
        if {$in ne "*"} {
          lappend row $in $xx($in)
	}
      }
      lappend rtn $row
    }
    set State(rows) $rtn
    if {$State(refSel) ne ""} {
      $State(refSel) configure -from 0 -to [expr {[llength $rtn]-1}]
      $State(refSel) set 3
      $State(refSel) set 0
    }
    my populate [lindex $State(rows) 0]
    return $rtn
  }

  
################################################################
#   method getSchema {tbl {save 1}}--
#    Return the schema for a table
# Arguments
#   tbl		Name of the table to get Schema for
#   save	if true, save this table as active
# Results
#   Modifies State(table) and State(primary) if save==1
# 
  method getSchema {tbl {save 1}} {
    variable State
    set sch [$State(db) eval "SELECT sql FROM sqlite_master WHERE name='$tbl'"]
    if {$save} {
      set State(table) $tbl
      set State(primary) [my getPrimary $tbl]
    }
    return $sch
  }
  
################################################################
#   method doSQL {sql}--
#    Evaluate an SQL string
# Arguments
#   sql		complete SQL command
# 
# Results
#   Returns results of SQL cmd
# 
  method doSQL {sql} {
    variable State
#    set x [$State(db) eval $sql]
#    return $x
    uplevel [list $State(db) eval $sql]
  }
  
################################################################
#   method getPrimary {tbl}--
#    Return the primary field for this table
# Arguments
#   tbl		Table to get primary for
# 
# Results
#   No side effects
# 
  method getPrimary {tbl} {
    variable State
# puts "XXXX"
# parray State
# puts "YYYY"

#    set sch [$State(db) eval "SELECT sql FROM sqlite_master WHERE name='$tbl'"]
    set sch [my getSchema $tbl 0]
    set lst [split $sch \n]

    set p [lsearch $lst "* primary *"]
    if {$p < 0} {
      puts "SQLITE_SCHEMA: $sch"; flush stdout
      error "Schema for $tbl has no primary key!"
    }
    return [lindex $lst $p 0]
  }

################################################################
#   method getValues {refList}--
#    Return a list of 
# Arguments
#   refList	List of fields for a reference
#    reference list would be things like tbl.field1 tbl.field2 that
#    are referenced by some primary key to tbl
# 
# Results
#   Modifies State
# 
  method getValues {refList} {
    variable State
    set rtn [list]
    
    if {[string first VALUES $refList] == 0} {
      return [lrange $refList 1 end]
    }

    foreach ref $refList {
# puts "REF: $ref"
      lassign [split $ref .] tbl fld

      set prim [my getPrimary $tbl]
      set State(primary,$ref) $prim
      
      set pos 0    
# puts "      $State(db) eval ''SELECT $prim, $fld FROM $tbl''" 
      $State(db) eval "SELECT $prim, $fld FROM $tbl" xx {
          #lappend State(ids,$ref) $xx($prim)
          lset rtn $pos [string trim "[lindex $rtn $pos] $xx($fld)"]
	  dict set State(idDict,$ref) [lindex $rtn $pos] $xx($prim)
          incr pos
      }
    }
    return $rtn
  }

################################################################
#   method doReplace {}--
#    Replace currently selected row with edited values
# Arguments
#   NONE
# 
# Results
#   Database is modified.
# 
  method doReplace {} {
    variable State
    set cmd {}
    foreach in [array names State value,*] {
      lassign [split $in ,] v fld
      if {[info exists State(combo,$fld)]} {
        #set $fld [lindex $State(ids,$State(ref,$fld)) [$State(combo,$fld) current]]
	set $fld [my getIDFromCombo $fld]
      } else {
        set $fld $State($in)
      }
# puts "FLD: $fld  EXIST [info exists State(post,$fld)]"
      if {[info exists State(post,$fld)]} {
        set $fld [$State(post,$fld) [set $fld]]
      }

      if {$fld ne $State(primary)} {
        append cmd "$fld = :$fld, "
      }
    }
    
    set cmd [string trim $cmd " ,"]
    
    set sql "UPDATE $State(table) SET $cmd WHERE $State(primary) = [set $State(primary)]"
    
    puts "SQL: $sql"
    $State(db) eval $sql
  }

################################################################
#   method doNew {}--
#    Add currently displayed values as a new row
# Arguments
#   NONE
# 
# Results
#   DB is modified
# 
  method doNew {} {
    variable State
# puts "NEW"
# parray State
# puts "NEW"
    set cmd {}
    set froms {}

    foreach in [array names State value,*] {
      lassign [split $in ,] v fld
      set dct [my getDict $fld]
# puts "IN: $in DCT: $dct"
      if {($dct ne "") && ([dict get $dct fromTbl] ne "")} {
        lappend froms $in
        continue
      }
      if {[info exists State(combo,$fld)]} {
        # set $fld [lindex $State(ids,$State(ref,$fld)) [$State(combo,$fld) current]]
	set $fld [my getIDFromCombo $fld]

      } else {
        set $fld $State($in)
      }

# puts "FLD: $fld  EXIST [info exists State(post,$fld)]"
      if {[info exists State(post,$fld)]} {
        set $fld [$State(post,$fld) [set $fld]]
      }

      if {![info exists State(primary)] || ($fld ne $State(primary))} {
        lappend flds $fld
	lappend vals :$fld
      }
    }
    
    set cmd [string trim $cmd " ,"]
    
    set sql "INSERT INTO $State(table) ([join $flds ,]) VALUES ([join $vals ,])"
    
     # puts "INSERT SQL: $sql"
    $State(db) eval $sql
    
    set new [$State(db) onecolumn "SELECT max($State(primary)) FROM $State(table)"]
    set State(value,$State(primary)) $new 
    
    foreach in $froms {
      lassign [split $in ,] v fld
      set dct [my getDict $fld]
# puts "IN: $in DCT: $dct"
      set fromVal $State($in)
      catch {unset tmp} 
      array set tmp $dct
      set fromFld $tmp(fromFld)
      set ln [my getSchemaLineRef [my getSchema $tmp(fromTbl) 0] $State(table)]
      # puts "LN: $ln"
      set idFld [lindex $ln 0]
      set flds [list $idFld $fromFld]
      set vals [list :new :from]
      foreach pair $tmp(fixed) {
        lassign $pair ff vv
        lappend flds $ff
	lappend vals :$ff
	set $ff $vv
      }
      set cmd "INSERT INTO $tmp(fromTbl) ([join $flds ,]) VALUES ([join $vals ,])"
      # puts "CMD: $cmd"
      $State(db) eval $cmd
    }
    # puts "DONE NEW"
  }
}

if {[info exists argv] && ([lsearch $argv -testET_SQLITE] >= 0)} {

lassign $argv tst testNum

catch {editTable destroy}
catch {destroy .t1}
source editTablesOO.tcl


source helpBalloon.tcl
package require tdbc::sqlite3
file delete test.db

set obj [editTable .t1 SQLITE3_support sqlite3 -dbArgs test.db]
set db [$obj config db]


$db eval {
  CREATE TABLE person (
    id integer unique primary key,
    loginid text, -- loginID --- {User Name} {LoginID} {} {-width 12 -bg green} {1 1}
    fname text, -- First name --- {First Name} {Given} {} {-width 12 -bg pink} {2 1}
    lname text, -- Last name --- {Last Name} {Surname} {} {-width 16 -bg yellow} {2 2}
    addrRef integer references addr -- Address Ref --- {} {Street Address} {addr.street} {-width 24} {3 1}
    );
  CREATE TABLE addr (
    id integer unique primary key,
    street text, -- Address --- {Street} {Addr1} {} {-width 26 -bg yellow} {3 1}
    city text, -- Address --- {CityHelp} {City} {} {-width 12 -bg yellow} {4 1}
    state text -- Address --- {StateHelp} {State} {} {-width 26 -bg yellow} 
  );
  CREATE TABLE phone ( 
    id integer unique primary key,
    type text,  --- {Home/Work} {Type} {person} {} {} {}
    number text, --- {Num} {Number} {person} {} {} {}
    owner_id integer references person --- {User} {Owner} {person.lname} {} {} $
    );
    

  CREATE TABLE usage (
    id integer unique primary key,
    useID text,  -- yes? No?
    useRedact text -- yes, no?
    );

  CREATE TABLE userphone ( 
    id integer unique primary key,
    type text,  --- {Home/Work} {Type} {person} {} {} {}
    number text, --- {Num} {Number} {person} {} {} {}
    owner_id integer references user --- {User} {Owner} {person.lname} {} {} $
    );
    
CREATE TABLE priv (
  id integer primary key,          -- unique ID
  name string                      --- {Privilege Name} {Privilege Name} {} {-width 30}
);

    
  CREATE TABLE useremail (
    id integer primary key,          -- unique ID
    userid integer references user,   --- {User Last Name} {User} {user.first_name user.last_name} {-width  20}
    type string,                      --- {Email Type(primary, work, sms, etc )} {Email} {user}
    address string                    --- {Address} {Address} {user} {}
  );

  CREATE TABLE user (
  id integer unique primary key,
  first_name string,           --- {First Name} {First Name} {} {-width 20} {1 0}
  middle_name string,          --- {Middle Name} {Middle Name} {} {-width 20} {1 1}
  last_name string,            --- {Last Name} {Last Name}  {} {-width 20} {1 2}
  city  string,                --- {City} {City} {} {-width 20} {6 0}
  state string,                --- {State} {State/Province} {} {-width 4} {6 1}
  language string,             --- {Language} {Language} {VALUES {} Italian Russian German {US English} Hindi {Portuguese (Brazilian)} {Spanish (Latin American)} {French Canadian}} {-width 8} {6 2}
  privilege integer references priv,   --- {Privilege Level} {Privileges} {priv.name} {} {2 0}
  userid string,               --- {Unique within site} {User ID} {} {-width 20} {2 1}
  login_name string,           --- {Login Name} {Login Name} {} {-width 20} {3 0}
  password string                  --- {Don't forget this password} {Password} {} {-width 20 -show *} {3 1} 
  );

}

$db eval {INSERT INTO addr VALUES (1, '123 St', 'Acity', 'AA')}
$db eval {INSERT INTO addr VALUES (2, '234 St', 'Bcity', 'BB')}
$db eval {INSERT INTO person VALUES (1, 'aaa', 'Alpha', 'Adam', 1)}
$db eval {INSERT INTO person VALUES (2, 'bbb', 'Beta', 'Blocker', 2)}
$db eval {INSERT INTO person VALUES (3, 'aa2', 'Abel', 'Adam', 2)}
$db eval {INSERT INTO person VALUES (4, 'aa3', 'Aard', 'Adam', 2)}
$db eval {INSERT INTO phone VALUES (1, 'work', '123-1234', '1')}
$db eval {INSERT INTO phone VALUES (2, 'work', '223-1234', '2')}
$db eval {INSERT INTO phone VALUES (3, 'home', '323-1234', '1')}
$db eval {INSERT INTO phone VALUES (4, 'home', '423-1234', '2')}
$db eval {INSERT INTO priv VALUES (1, 'user')}
$db eval {INSERT INTO priv VALUES (2, 'superuser')}

if {$testNum == 1} {
  set obj2 [.t1 clone .t2]
  .t2,frame configure -borderwidth 2 -relief solid

  $obj makeGUI [$obj getSchema person]
  $obj config -table person
  $obj populateBySearch "loginid = 'aaa'"
  
  $obj2 makeGUI [$obj2 getSchema addr]
  $obj2 config -table addr
  $obj2 populateBySearch "city= 'Bcity'"
  grid .t1
  grid .t2
}

if {$testNum == 2} {
  toplevel .tt
  set obj2 [.t1 clone .tt.t2]
  $obj2 makeGUI [$obj2 getSchema addr]
  $obj2 config -table addr
  $obj2 populateBySearch "city= 'Bcity'"
  grid .tt.t2
 }

if {$testNum == 3} {
  toplevel .tt
  set obj2 [.t1 clone .tt.t2]
  $obj2 configure -table usage

  set dicts {
    {primary id fld useID help "Use ID?" label "Use ID" values {yes no} ref {} args {} gridOpts {}}
    {fld useRedact help "Use Redact?" label "Use Redact" values {yes no} ref {} args {} gridOpts {}}
  }
  $obj2 makeGUIwDict $dicts full

  grid .tt.t2
 }

if {$testNum == 4} {
  toplevel .tt
  set obj2 [.t1 clone .tt.t2]

  set sch { TABLE CREATE user ( 
  id integer unique primary key,
  first_name string,           --- {First Name} {First Name} {} {-width 20} {1 0}
  middle_name string,          --- {Middle Name} {Middle Name} {} {-width 20} {1 1}
  last_name string,            --- {Last Name} {Last Name}  {} {-width 20} {1 2}
  city  string,                --- {City} {City} {} {-width 20} {6 0}
  state string,                --- {State} {State/Province} {} {-width 4} {6 1}
  language string,             --- {Language} {Language} {VALUES {} Italian Russian German {US English} Hindi {Portuguese (Brazilian)} {Spanish (Latin American)} {French Canadian}} {-width 8} {6 2}
  work_email string,           --- {Work Email} {Work Email} {useremail.address} {} {4 0}
  personal_email string,       --- {Personal Email} {Personal Email} {useremail.address} {} {4 1}
  work_phone string,               --- {Work Phone} {Work Phone} {userphone.number} {} {5 0}
  home_phone string,               --- {Home Phone} {Home Phone} {userphone.number} {} {5 1}
  privilege integer references priv,   --- {Privilege Level} {Privileges} {priv.name} {} {2 0}
  userid string,               --- {Unique within site} {User ID} {} {-width 20} {2 1}
  login_name string,           --- {Login Name} {Login Name} {} {-width 20} {3 0}
  password string                  --- {Don't forget this password} {Password} {} {-width 20 -show *} {3 1} 
  )}


  $obj2 makeGUI $sch
  set dd [$obj2 getDict work_phone]
  # puts "DD: $dd"
  dict set dd fixed [list [list type work]]
  $obj2 setDict work_phone $dd
  grid .tt.t2
 }

}



